@@ -21,9 +21,19 @@
 #include <stdlib.h>
 #include <gtk/gtk.h>
 #include <gdk/gdk.h>
+#include <gdk/gdkx.h>
 #include <glib/gi18n.h>
 #include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+#include <limits.h>
 #include <signal.h>
+#include <stdlib.h>
+#include <string.h>
+#include <unistd.h>
+
+#include <X11/Xatom.h>
+#include <X11/Xlib.h>
 
 #include "dbus-interface.h"
 
@@ -57,10 +67,12 @@ typedef struct {
     int reboot_HAL : 1;			/* Reboot is available via HAL */
     int suspend_HAL : 1;		/* Suspend is available via HAL */
     int hibernate_HAL : 1;		/* Hibernate is available via HAL */
-    int switch_user_KDE : 1;		/* Switch User is available via KDE */
-
+    int switch_user_GDM : 1;		/* Switch User is available via GDM */
+    int switch_user_KDM : 1;		/* Switch User is available via KDM */
+    int ltsp : 1; /* under LTSP environment */
 } HandlerContext;
 
+static gboolean verify_running(char * display_manager, char * executable);
 static void logout_clicked(GtkButton * button, HandlerContext * handler_context);
 static void shutdown_clicked(GtkButton * button, HandlerContext * handler_context);
 static void reboot_clicked(GtkButton * button, HandlerContext * handler_context);
@@ -71,6 +83,61 @@ static void cancel_clicked(GtkButton * button, gpointer user_data);
 static GtkPositionType get_banner_position(void);
 static GdkPixbuf * get_background_pixbuf(void);
 
+/* Verify that a program is running and that an executable is available. */
+static gboolean verify_running(char * display_manager, char * executable)
+{
+    /* See if the executable we need to run is in the path. */
+    gchar * full_path = g_find_program_in_path(executable);
+    if (full_path != NULL)
+    {
+        g_free(full_path);
+
+        /* Form the filespec of the pid file for the display manager. */
+        char buffer[PATH_MAX];
+        sprintf(buffer, "/var/run/%s.pid", display_manager);
+
+        /* Open the pid file. */
+        int fd = open(buffer, O_RDONLY);
+        if (fd >= 0)
+        {
+            /* Pid file exists.  Read it. */
+            ssize_t length = read(fd, buffer, sizeof(buffer));
+            close(fd);
+            if (length > 0)
+            {
+                /* Null terminate the buffer and convert the pid. */
+                buffer[length] = '\0';
+                pid_t pid = atoi(buffer);
+                if (pid > 0)
+                {
+                    /* Form the filespec of the command line file under /proc.
+                     * This is Linux specific.  Should be conditionalized to the appropriate /proc layout for
+                     * other systems.  Your humble developer has no way to test on other systems. */
+                    sprintf(buffer, "/proc/%d/cmdline", pid);
+
+                    /* Open the file. */
+                    int fd = open(buffer, O_RDONLY);
+                    if (fd >= 0)
+                    {
+                        /* Read the command line. */
+                        ssize_t length = read(fd, buffer, sizeof(buffer));
+                        close(fd);
+                        if (length > 0)
+                        {
+                            /* Null terminate the buffer and look for the display manager name in the command.
+                             * If found, return success. */
+                            buffer[length] = '\0';
+                            if (strstr(buffer, display_manager) != NULL)
+                                return TRUE;
+                        }
+                    }
+                }
+            }
+        }
+    }
+    return FALSE;
+}
+
 /* Handler for "clicked" signal on Logout button. */
 static void logout_clicked(GtkButton * button, HandlerContext * handler_context)
 {
@@ -78,10 +145,24 @@ static void logout_clicked(GtkButton * button, HandlerContext * handler_context)
     gtk_main_quit();
 }
 
+static void change_root_property(GtkWidget* w, const char* prop_name, const char* value)
+{
+    GdkDisplay* dpy = gtk_widget_get_display(w);
+    GdkWindow* root = gtk_widget_get_root_window(w);
+    XChangeProperty(GDK_DISPLAY_XDISPLAY(dpy), GDK_WINDOW_XID(root),
+                      XInternAtom(GDK_DISPLAY_XDISPLAY(dpy), prop_name, False), XA_STRING, 8,
+                      PropModeReplace, value, strlen(value) + 1);
+}
+
 /* Handler for "clicked" signal on Shutdown button. */
 static void shutdown_clicked(GtkButton * button, HandlerContext * handler_context)
 {
-    if (handler_context->shutdown_ConsoleKit)
+    if (G_UNLIKELY(handler_context->ltsp))
+    {
+        change_root_property(GTK_WIDGET(button), "LTSP_LOGOUT_ACTION", "HALT");
+        kill(handler_context->lxsession_pid, SIGTERM);
+    }
+    else if (handler_context->shutdown_ConsoleKit)
         dbus_ConsoleKit_Stop();
     else if (handler_context->shutdown_HAL)
         dbus_HAL_Shutdown();
@@ -91,7 +172,12 @@ static void shutdown_clicked(GtkButton * button, HandlerContext * handler_contex
 /* Handler for "clicked" signal on Reboot button. */
 static void reboot_clicked(GtkButton * button, HandlerContext * handler_context)
 {
-    if (handler_context->reboot_ConsoleKit)
+    if (G_UNLIKELY(handler_context->ltsp))
+    {
+        change_root_property(GTK_WIDGET(button), "LTSP_LOGOUT_ACTION", "REBOOT");
+        kill(handler_context->lxsession_pid, SIGTERM);
+    }
+    else if (handler_context->reboot_ConsoleKit)
         dbus_ConsoleKit_Restart();
     else if (handler_context->reboot_HAL)
         dbus_HAL_Reboot();
@@ -121,7 +207,9 @@ static void hibernate_clicked(GtkButton * button, HandlerContext * handler_conte
 /* Handler for "clicked" signal on Switch User button. */
 static void switch_user_clicked(GtkButton * button, HandlerContext * handler_context)
 {
-    if (handler_context->switch_user_KDE)
+    if (handler_context->switch_user_GDM)
+        g_spawn_command_line_sync("gdmflexiserver --startnew", NULL, NULL, NULL, NULL);
+    else if (handler_context->switch_user_KDM)
         g_spawn_command_line_sync("kdmctl reserve", NULL, NULL, NULL, NULL);
     gtk_main_quit();
 }
@@ -289,18 +377,24 @@ int main(int argc, char * argv[])
         handler_context.hibernate_HAL = TRUE;
     }
 
-    /* If we are under KDM, its "Switch User" is available. */
-    if (g_file_test("/var/run/kdm.pid", G_FILE_TEST_EXISTS))
+    /* If we are under GDM, its "Switch User" is available. */
+    if (verify_running("gdm", "gdmflexiserver"))
     {
-        gchar * test = g_find_program_in_path("kdmctl");
-        if (test != NULL)
-        {
-            g_free(test);
-            handler_context.switch_user_available = TRUE;
-            handler_context.switch_user_KDE = TRUE;
-        }
+        handler_context.switch_user_available = TRUE;
+        handler_context.switch_user_GDM = TRUE;
     }
 
+    /* If we are under KDM, its "Switch User" is available. */
+    if (verify_running("kdm", "kdmctl"))
+    {
+        handler_context.switch_user_available = TRUE;
+        handler_context.switch_user_KDM = TRUE;
+    }
+
+    /* LTSP support */
+    if (g_getenv("LTSP_CLIENT"))
+        handler_context.ltsp = TRUE;
+
     /* Make the button images accessible. */
     gtk_icon_theme_append_search_path(gtk_icon_theme_get_default(), PACKAGE_DATA_DIR "/lxsession/images");
 
@@ -395,6 +489,7 @@ int main(int argc, char * argv[])
         GtkWidget * shutdown_button = gtk_button_new_with_mnemonic(_("Sh_utdown"));
         GtkWidget * image = gtk_image_new_from_icon_name("system-shutdown", GTK_ICON_SIZE_BUTTON);
         gtk_button_set_image(GTK_BUTTON(shutdown_button), image);
+        gtk_button_set_alignment(GTK_BUTTON(shutdown_button), 0.0, 0.5);
         g_signal_connect(G_OBJECT(shutdown_button), "clicked", G_CALLBACK(shutdown_clicked), &handler_context);
         gtk_box_pack_start(GTK_BOX(controls), shutdown_button, FALSE, FALSE, 4);
     }
@@ -405,36 +500,40 @@ int main(int argc, char * argv[])
         GtkWidget * reboot_button = gtk_button_new_with_mnemonic(_("_Reboot"));
         GtkWidget * image = gtk_image_new_from_icon_name("gnome-session-reboot", GTK_ICON_SIZE_BUTTON);
         gtk_button_set_image(GTK_BUTTON(reboot_button), image);
+        gtk_button_set_alignment(GTK_BUTTON(reboot_button), 0.0, 0.5);
         g_signal_connect(G_OBJECT(reboot_button), "clicked", G_CALLBACK(reboot_clicked), &handler_context);
         gtk_box_pack_start(GTK_BOX(controls), reboot_button, FALSE, FALSE, 4);
     }
 
     /* Create the Suspend button. */
-    if (handler_context.suspend_available)
+    if (handler_context.suspend_available && !handler_context.ltsp)
     {
         GtkWidget * suspend_button = gtk_button_new_with_mnemonic(_("_Suspend"));
         GtkWidget * image = gtk_image_new_from_icon_name("gnome-session-suspend", GTK_ICON_SIZE_BUTTON);
         gtk_button_set_image(GTK_BUTTON(suspend_button), image);
+        gtk_button_set_alignment(GTK_BUTTON(suspend_button), 0.0, 0.5);
         g_signal_connect(G_OBJECT(suspend_button), "clicked", G_CALLBACK(suspend_clicked), &handler_context);
         gtk_box_pack_start(GTK_BOX(controls), suspend_button, FALSE, FALSE, 4);
     }
 
     /* Create the Hibernate button. */
-    if (handler_context.hibernate_available)
+    if (handler_context.hibernate_available && !handler_context.ltsp)
     {
         GtkWidget * hibernate_button = gtk_button_new_with_mnemonic(_("_Hibernate"));
         GtkWidget * image = gtk_image_new_from_icon_name("gnome-session-hibernate", GTK_ICON_SIZE_BUTTON);
         gtk_button_set_image(GTK_BUTTON(hibernate_button), image);
+        gtk_button_set_alignment(GTK_BUTTON(hibernate_button), 0.0, 0.5);
         g_signal_connect(G_OBJECT(hibernate_button), "clicked", G_CALLBACK(hibernate_clicked), &handler_context);
         gtk_box_pack_start(GTK_BOX(controls), hibernate_button, FALSE, FALSE, 4);
     }
 
     /* Create the Switch User button. */
-    if (handler_context.switch_user_available)
+    if (handler_context.switch_user_available && !handler_context.ltsp)
     {
         GtkWidget * switch_user_button = gtk_button_new_with_mnemonic(_("S_witch User"));
         GtkWidget * image = gtk_image_new_from_icon_name("gnome-session-switch", GTK_ICON_SIZE_BUTTON);
         gtk_button_set_image(GTK_BUTTON(switch_user_button), image);
+        gtk_button_set_alignment(GTK_BUTTON(switch_user_button), 0.0, 0.5);
         g_signal_connect(G_OBJECT(switch_user_button), "clicked", G_CALLBACK(switch_user_clicked), &handler_context);
         gtk_box_pack_start(GTK_BOX(controls), switch_user_button, FALSE, FALSE, 4);
     }
@@ -443,11 +542,13 @@ int main(int argc, char * argv[])
     GtkWidget * logout_button = gtk_button_new_with_mnemonic(_("_Logout"));
     GtkWidget * image = gtk_image_new_from_icon_name("system-log-out", GTK_ICON_SIZE_BUTTON);
     gtk_button_set_image(GTK_BUTTON(logout_button), image);
+    gtk_button_set_alignment(GTK_BUTTON(logout_button), 0.0, 0.5);
     g_signal_connect(G_OBJECT(logout_button), "clicked", G_CALLBACK(logout_clicked), &handler_context);
     gtk_box_pack_start(GTK_BOX(controls), logout_button, FALSE, FALSE, 4);
 
     /* Create the Cancel button. */
     GtkWidget * cancel_button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
+    gtk_button_set_alignment(GTK_BUTTON(cancel_button), 0.0, 0.5);
     g_signal_connect(G_OBJECT(cancel_button), "clicked", G_CALLBACK(cancel_clicked), NULL);
     gtk_box_pack_start(GTK_BOX(controls), cancel_button, FALSE, FALSE, 4);
 
